# -*- coding: utf-8 -*-
#
# Copyright (C) 2025 CERN.
#
# chATLAS_Embed is free software; you can redistribute it and/or modify
# it under the terms of the Apache 2.0 license; see LICENSE file for more details.

"""
`chATLAS_Embed/LangChainVectorStore.py`

Wrapper for vector store to be compatible with langchain
"""

from .Base import VectorStore
import asyncio

from langchain_core.callbacks import CallbackManagerForRetrieverRun
from langchain_core.documents import Document as LangchainDocument
from langchain_core.retrievers import BaseRetriever
from pydantic import BaseModel, Field
from typing import List

class LangChainVectorStore(BaseRetriever, BaseModel):
    """
    Adapter for using a custom VectorStore store with LangChain.
    **NOTE**: Not all methods and features of langchain/langgraph currently supported.

    Inputs:
    vector_store: VectorStore - Postgres vectorstore
    """

    vector_store: VectorStore = Field(..., description="The underlying vector store instance.")

    class Config:
        arbitrary_types_allowed = True  # Allow non-Pydantic types for vector_store

    def search(self, query: str, k: int = 4, metadata_filters: dict = None) -> List[LangchainDocument]:
        """
        Search and return documents in LangChain-compatible format.

        :param query: (str) - Input query to search vectorstore with
        :param k: (int) - Number of documents to return
        :param metadata_filters: (dict) - Dictionary of metadata filters where keys are metadata fields and values are filter
            values.
            can contain key: date_filter (str): A date string (e.g., '2010-04') to filter 'last_modification' for newer entries.

        :return:
        Search Results: (List[LangchainDocument])
        """

        # Extract date_filter from metadata filters
        date_filter = metadata_filters.get("date_filter", None)
        if date_filter:
            del metadata_filters["date_filter"]

        # Extract k value from metadata filters
        k_new = metadata_filters.get("k", None)
        if k_new:
            k = k_new
            del metadata_filters["k"]

        results = self.vector_store.search(query, k, metadata_filters=metadata_filters,
                                           date_filter=date_filter)  # Returns List[(child, parent, score)]
        formatted_results = []

        for child_doc, parent_doc, similarity in results:
            # Combine child and parent metadata for context if needed

            # Adding just parent metadata
            metadata = {}
            for key, val in parent_doc.metadata.items():
                metadata[key] = val
            metadata["similarity"] = similarity

            langchain_formatted_doc = LangchainDocument(
                page_content=parent_doc.page_content,
                metadata=metadata
            )
            formatted_results.append(langchain_formatted_doc)

        return formatted_results

    async def _aget_relevant_documents(self, query: str, run_manager: CallbackManagerForRetrieverRun = None) -> List[
        LangchainDocument]:
        """
        Retrieve relevant documents asynchronously.

        :param query: (str) - Input query to search vectorstore with
        :param run_manager: (CallbackManagerForRetrieverRun) - Autogenerated and passed in by langchain

        :return:
        Search Results: (List[LangchainDocument])
        """

        # Clean metadata from langchain rubbish
        input_metadata = {k: v for k, v in (run_manager.metadata.items() if run_manager else {}).items() if
                          not k.startswith("ls_")}

        # Define a wrapper for `self.search`
        def search_wrapper():
            return self.search(query=query, k=4, metadata_filters=input_metadata)

        return await asyncio.to_thread(search_wrapper)

    def _get_relevant_documents(self, query: str, *, run_manager: CallbackManagerForRetrieverRun) -> list[
        LangchainDocument]:
        """
        Retrieve relevant documents asynchronously.

        :param query: (str) - Input query to search vectorstore with
        :param run_manager: (CallbackManagerForRetrieverRun) - Autogenerated and passed in by langchain

        :return:
        Search Results: (List[LangchainDocument])
        """
        input_metadata = {k: v for k, v in run_manager.metadata.items() if not k.startswith("ls_")}
        return self.search(query, k=4, metadata_filters=input_metadata)
